class SPHCompiler(object):
    def __init__(self, acceleration_evals, integrator):
        """Compiles the acceleration evaluator and integrator to produce a
        fast version using one of the supported backends. If the backend is
        not given, one is automatically chosen based on the configuration.

        Parameters
        ----------

        acceleration_eval: .acceleration_eval.AccelerationEval instance
            or list of instances.
        integrator: .integrator.Integrator instance
        """
        if not isinstance(acceleration_evals, (list, tuple)):
            acceleration_evals = [acceleration_evals]

        self.acceleration_evals = list(acceleration_evals)
        self.integrator = integrator
        if integrator is not None:
            integrator.set_acceleration_evals(self.acceleration_evals)
        self.backend = acceleration_evals[0].backend
        self._setup_helpers()
        self.module = None

    # Public interface. ####################################################
    def compile(self):
        """Compile the generated code to extension modules and
        setup the objects that need this by calling their
        setup_compiled_module.
        """
        if self.module is not None:
            return

        # We compile the first acceleration eval along with the integrator.
        # The rest of the acceleration evals (if present) are independent.
        code0 = self._get_code()
        helper0 = self.acceleration_eval_helpers[0]
        mod = helper0.compile(code0)
        helper0.setup_compiled_module(mod)

        self.module = mod
        c_accel_eval0 = self.acceleration_evals[0].c_acceleration_eval

        if self.backend == 'cython':
            if self.integrator is not None:
                self.integrator_helper.setup_compiled_module(
                    mod, c_accel_eval0
                )
        elif self.backend == 'opencl' or self.backend == 'cuda':
            if self.integrator is not None:
                self.integrator_helper.setup_compiled_module(
                    mod, c_accel_eval0
                )

        # Setup the remaining acceleration evals.
        for helper in self.acceleration_eval_helpers[1:]:
            mod = helper.compile(helper.get_code())
            helper.setup_compiled_module(mod)

    # Private interface. ####################################################
    def _get_code(self):
        main = self.acceleration_eval_helpers[0].get_code()
        integrator_code = self.integrator_helper.get_code()
        return main + integrator_code

    def _setup_helpers(self):
        if self.backend == 'cython':
            from .acceleration_eval_cython_helper import \
                AccelerationEvalCythonHelper
            cls = AccelerationEvalCythonHelper
        elif self.backend == 'opencl' or self.backend == 'cuda':
            from .acceleration_eval_gpu_helper import \
                AccelerationEvalGPUHelper
            cls = AccelerationEvalGPUHelper

        self.acceleration_eval_helpers = [
            cls(a_eval)
            for a_eval in self.acceleration_evals
        ]
        self._setup_integrator_helper()

    def _setup_integrator_helper(self):
        a_helper0 = self.acceleration_eval_helpers[0]
        if self.backend == 'cython':
            from .integrator_cython_helper import \
                IntegratorCythonHelper
            self.integrator_helper = IntegratorCythonHelper(
                self.integrator, a_helper0
            )
        elif self.backend == 'opencl' or self.backend == 'cuda':
            from .integrator_gpu_helper import IntegratorGPUHelper
            self.integrator_helper = IntegratorGPUHelper(
                self.integrator, a_helper0
            )



"""Integrator steps for different schemes.

Implement as many stages as needed.
"""


###############################################################################
# `IntegratorStep` class
###############################################################################
class IntegratorStep(object):
    """Subclass this and implement the methods ``initialize``, ``stage1`` etc.
    Use the same conventions as the equations.
    """
    def __repr__(self):
        return '%s()'%(self.__class__.__name__)


###############################################################################
# `EulerStep` class
###############################################################################
class EulerStep(IntegratorStep):
    """Fast but inaccurate integrator. Use this for testing"""
    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_x, d_y,
                  d_z, d_rho, d_arho, dt):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

        d_x[d_idx] += dt*d_u[d_idx]
        d_y[d_idx] += dt*d_v[d_idx]
        d_z[d_idx] += dt*d_w[d_idx]

        d_rho[d_idx] += dt*d_arho[d_idx]

###############################################################################
# `WCSPHStep` class
###############################################################################
class WCSPHStep(IntegratorStep):
    """Standard Predictor Corrector integrator for the WCSPH formulation

    Use this integrator for WCSPH formulations. In the predictor step,
    the particles are advanced to `t + dt/2`. The particles are then
    advanced with the new force computed at this position.

    This integrator can be used in PEC or EPEC mode.

    The same integrator can be used for other problems. Like for
    example solid mechanics (see SolidMechStep)

    """
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
                   d_aw, d_ax, d_ay, d_az, d_arho, dt):
        dtb2 = 0.5*dt
        d_u[d_idx] = d_u0[d_idx] + dtb2*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_az[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_rho[d_idx] = d_rho0[d_idx] + dtb2 * d_arho[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
                   d_aw, d_ax, d_ay, d_az, d_arho, dt):

        d_u[d_idx] = d_u0[d_idx] + dt*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_az[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_rho[d_idx] = d_rho0[d_idx] + dt * d_arho[d_idx]

###############################################################################
# `WCSPHTVDRK3` Integrator
###############################################################################
class WCSPHTVDRK3Step(IntegratorStep):
    r"""TVD RK3 stepper for WCSPH

    This integrator requires :math:`2` stages for the storage of the
    acceleration variables.

    """
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho,
               d_au, d_av, d_aw, d_ax, d_ay, d_az, d_arho,
               dt):

        # update velocities
        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        # update positions
        d_x[d_idx] = d_x0[d_idx] + dt * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_az[d_idx]

        # update density
        d_rho[d_idx] = d_rho0[d_idx] + dt * d_arho[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
               d_aw, d_ax, d_ay, d_az, d_arho, dt):

        # update velocities
        d_u[d_idx] = 0.75*d_u0[d_idx] + 0.25*( d_u[d_idx] + dt * d_au[d_idx] )
        d_v[d_idx] = 0.75*d_v0[d_idx] + 0.25*( d_v[d_idx] + dt * d_av[d_idx] )
        d_w[d_idx] = 0.75*d_w0[d_idx] + 0.25*( d_w[d_idx] + dt * d_aw[d_idx] )

        # update positions
        d_x[d_idx] = 0.75*d_x0[d_idx] + 0.25*( d_x[d_idx] + dt * d_ax[d_idx] )
        d_y[d_idx] = 0.75*d_y0[d_idx] + 0.25*( d_y[d_idx] + dt * d_ay[d_idx] )
        d_z[d_idx] = 0.75*d_z0[d_idx] + 0.25*( d_z[d_idx] + dt * d_az[d_idx] )

        # Update density
        d_rho[d_idx] = 0.75*d_rho0[d_idx] + 0.25*( d_rho[d_idx] + dt * d_arho[d_idx] )

    def stage3(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
               d_aw, d_ax, d_ay, d_az, d_arho, dt):

        oneby3 = 1./3.
        twoby3 = 2./3.

        # update velocities
        d_u[d_idx] = oneby3*d_u0[d_idx] + twoby3*( d_u[d_idx] + dt * d_au[d_idx] )
        d_v[d_idx] = oneby3*d_v0[d_idx] + twoby3*( d_v[d_idx] + dt * d_av[d_idx] )
        d_w[d_idx] = oneby3*d_w0[d_idx] + twoby3*( d_w[d_idx] + dt * d_aw[d_idx] )

        # update positions
        d_x[d_idx] = oneby3*d_x0[d_idx] + twoby3*( d_x[d_idx] + dt * d_ax[d_idx] )
        d_y[d_idx] = oneby3*d_y0[d_idx] + twoby3*( d_y[d_idx] + dt * d_ay[d_idx] )
        d_z[d_idx] = oneby3*d_z0[d_idx] + twoby3*( d_z[d_idx] + dt * d_az[d_idx] )

        # Update density
        d_rho[d_idx] = oneby3*d_rho0[d_idx] + twoby3*( d_rho[d_idx] + dt * d_arho[d_idx] )

###############################################################################
# `SolidMechStep` class
###############################################################################
class SolidMechStep(IntegratorStep):
    """Predictor corrector Integrator for solid mechanics problems"""
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho,
                   d_s00, d_s01, d_s02, d_s11, d_s12, d_s22,
                   d_s000, d_s010, d_s020, d_s110, d_s120, d_s220,
                   d_e0, d_e):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]
        d_e0[d_idx] = d_e[d_idx]

        d_s000[d_idx] = d_s00[d_idx]
        d_s010[d_idx] = d_s01[d_idx]
        d_s020[d_idx] = d_s02[d_idx]
        d_s110[d_idx] = d_s11[d_idx]
        d_s120[d_idx] = d_s12[d_idx]
        d_s220[d_idx] = d_s22[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                  d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
                  d_aw, d_ax, d_ay, d_az, d_arho, d_e, d_e0, d_ae,
                  d_s00, d_s01, d_s02, d_s11, d_s12, d_s22,
                  d_s000, d_s010, d_s020, d_s110, d_s120, d_s220,
                  d_as00, d_as01, d_as02, d_as11, d_as12, d_as22,
                  dt):
        dtb2 = 0.5*dt
        d_u[d_idx] = d_u0[d_idx] + dtb2*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_az[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_rho[d_idx] = d_rho0[d_idx] + dtb2 * d_arho[d_idx]
        d_e[d_idx] = d_e0[d_idx] + dtb2 * d_ae[d_idx]

        # update deviatoric stress components
        d_s00[d_idx] = d_s000[d_idx] + dtb2 * d_as00[d_idx]
        d_s01[d_idx] = d_s010[d_idx] + dtb2 * d_as01[d_idx]
        d_s02[d_idx] = d_s020[d_idx] + dtb2 * d_as02[d_idx]
        d_s11[d_idx] = d_s110[d_idx] + dtb2 * d_as11[d_idx]
        d_s12[d_idx] = d_s120[d_idx] + dtb2 * d_as12[d_idx]
        d_s22[d_idx] = d_s220[d_idx] + dtb2 * d_as22[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                  d_u0, d_v0, d_w0, d_u, d_v, d_w, d_rho0, d_rho, d_au, d_av,
                  d_aw, d_ax, d_ay, d_az, d_arho, d_e, d_ae, d_e0,
                  d_s00, d_s01, d_s02, d_s11, d_s12, d_s22,
                  d_s000, d_s010, d_s020, d_s110, d_s120, d_s220,
                  d_as00, d_as01, d_as02, d_as11, d_as12, d_as22,
                  dt):

        d_u[d_idx] = d_u0[d_idx] + dt*d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt*d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_ax[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_ay[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_az[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_rho[d_idx] = d_rho0[d_idx] + dt * d_arho[d_idx]
        d_e[d_idx] = d_e0[d_idx] + dt * d_ae[d_idx]

        # update deviatoric stress components
        d_s00[d_idx] = d_s000[d_idx] + dt * d_as00[d_idx]
        d_s01[d_idx] = d_s010[d_idx] + dt * d_as01[d_idx]
        d_s02[d_idx] = d_s020[d_idx] + dt * d_as02[d_idx]
        d_s11[d_idx] = d_s110[d_idx] + dt * d_as11[d_idx]
        d_s12[d_idx] = d_s120[d_idx] + dt * d_as12[d_idx]
        d_s22[d_idx] = d_s220[d_idx] + dt * d_as22[d_idx]

###############################################################################
# `TransportVelocityStep` class
###############################################################################
class TransportVelocityStep(IntegratorStep):
    """Integrator defined in 'A transport velocity formulation for
    smoothed particle hydrodynamics', 2013, JCP, 241, pp 292--307

    For a predictor-corrector style of integrator, this integrator
    should operate only in PEC mode.

    """
    def initialize(self):
        pass

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_uhat, d_auhat, d_vhat,
                  d_avhat, d_what, d_awhat, d_x, d_y, d_z, dt):
        dtb2 = 0.5*dt

        # velocity update eqn (14)
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]

        # advection velocity update eqn (15)
        d_uhat[d_idx] = d_u[d_idx] + dtb2*d_auhat[d_idx]
        d_vhat[d_idx] = d_v[d_idx] + dtb2*d_avhat[d_idx]
        d_what[d_idx] = d_w[d_idx] + dtb2*d_awhat[d_idx]

        # position update eqn (16)
        d_x[d_idx] += dt*d_uhat[d_idx]
        d_y[d_idx] += dt*d_vhat[d_idx]
        d_z[d_idx] += dt*d_what[d_idx]

    def stage2(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_vmag2, dt):
        dtb2 = 0.5*dt

        # corrector update eqn (17)
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]

        # magnitude of velocity squared
        d_vmag2[d_idx] = (d_u[d_idx]*d_u[d_idx] + d_v[d_idx]*d_v[d_idx] +
                          d_w[d_idx]*d_w[d_idx])

###############################################################################
# `AdamiVerletStep` class
###############################################################################
class AdamiVerletStep(IntegratorStep):
    """Verlet time integration described in `A generalized wall
    boundary condition for smoothed particle hydrodynamics` 2012, JCP,
    231, pp 7057--7075

    This integrator can operate in either PEC mode or in EPEC mode as
    described in the paper.

    """
    def initialize(self):
        pass

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_x, d_y, d_z, dt):
        dtb2 = 0.5*dt

        # velocity predictor eqn (14)
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]

        # position predictor eqn (15)
        d_x[d_idx] += dtb2*d_u[d_idx]
        d_y[d_idx] += dtb2*d_v[d_idx]
        d_z[d_idx] += dtb2*d_w[d_idx]

    def stage2(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, d_x, d_y, d_z,
               d_rho, d_arho, d_vmag2, dt):
        dtb2 = 0.5*dt

        # position corrector eqn (17)
        d_x[d_idx] += dtb2*d_u[d_idx]
        d_y[d_idx] += dtb2*d_v[d_idx]
        d_z[d_idx] += dtb2*d_w[d_idx]

        # velocity corrector eqn (18)
        d_u[d_idx] += dtb2*d_au[d_idx]
        d_v[d_idx] += dtb2*d_av[d_idx]
        d_w[d_idx] += dtb2*d_aw[d_idx]

        # density corrector eqn (16)
        d_rho[d_idx] += dt * d_arho[d_idx]

        # magnitude of velocity squared
        d_vmag2[d_idx] = (d_u[d_idx]*d_u[d_idx] + d_v[d_idx]*d_v[d_idx] +
                          d_w[d_idx]*d_w[d_idx])

###############################################################################
# `GasDFluidStep` class
###############################################################################
class GasDFluidStep(IntegratorStep):
    """Predictor Corrector integrator for Gas-dynamics"""
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_h,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e, d_e0, d_h0,
                   d_converged, d_omega, d_rho, d_rho0, d_alpha1, d_alpha2,
                   d_alpha10, d_alpha20):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_e0[d_idx] = d_e[d_idx]

        d_h0[d_idx] = d_h[d_idx]
        d_rho0[d_idx] = d_rho[d_idx]

        # set the converged attribute to 0 at the beginning of a Group
        d_converged[d_idx] = 0

        # likewise, we set the default omega (grad-h) terms to 1 at
        # the beginning of this Group.
        d_omega[d_idx] = 1.0

        d_alpha10[d_idx] = d_alpha1[d_idx]
        d_alpha20[d_idx] = d_alpha2[d_idx]

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e0, d_e, d_au, d_av,
               d_aw, d_ae, d_rho, d_rho0, d_arho, d_h, d_h0, d_ah,
               d_alpha1, d_aalpha1, d_alpha10,
               d_alpha2, d_aalpha2, d_alpha20,
               dt):
        dtb2 = 0.5*dt

        d_u[d_idx] = d_u0[d_idx] + dtb2 * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2 * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2 * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_w[d_idx]

        # update thermal energy
        d_e[d_idx] = d_e0[d_idx] + dtb2 * d_ae[d_idx]

        # predict density and smoothing lengths for faster
        # convergence. NNPS need not be explicitly updated since it
        # will be called at the end of the predictor stage.
        d_h[d_idx] = d_h0[d_idx] + dtb2 * d_ah[d_idx]
        d_rho[d_idx] = d_rho0[d_idx] + dtb2 * d_arho[d_idx]

        # update viscosity coefficients
        d_alpha1[d_idx] = d_alpha10[d_idx] + dtb2*d_aalpha1[d_idx]
        d_alpha2[d_idx] = d_alpha20[d_idx] + dtb2*d_aalpha2[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e0, d_e, d_au, d_av,
               d_alpha1, d_aalpha1, d_alpha10,
               d_alpha2, d_aalpha2, d_alpha20,
               d_aw, d_ae, dt):

        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_e[d_idx] = d_e0[d_idx] + dt * d_ae[d_idx]

        # update viscosity coefficients
        d_alpha1[d_idx] = d_alpha10[d_idx] + dt*d_aalpha1[d_idx]
        d_alpha2[d_idx] = d_alpha20[d_idx] + dt*d_aalpha2[d_idx]


class GSPHStep(IntegratorStep):

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_e,
               d_au, d_av, d_aw, d_ae, dt):
        dtb2 = dt*0.5
        ustar = d_u[d_idx] + dtb2*d_au[d_idx]
        vstar = d_v[d_idx] + dtb2*d_av[d_idx]
        wstar = d_w[d_idx] + dtb2*d_aw[d_idx]

        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]
        d_e[d_idx] += dt*(d_ae[d_idx]
                          - ustar*d_au[d_idx]
                          - vstar*d_av[d_idx]
                          - wstar*d_aw[d_idx])
        d_x[d_idx] += dt*ustar
        d_y[d_idx] += dt*vstar
        d_z[d_idx] += dt*wstar


class ADKEStep(IntegratorStep):
    """Predictor Corrector integrator for Gas-dynamics ADKE"""
    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
                   d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e, d_e0,
                   d_rho, d_rho0):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_e0[d_idx] = d_e[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]


    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e0, d_e, d_au, d_av,
               d_aw, d_ae, d_rho, d_rho0, d_arho, dt):
        dtb2 = 0.5*dt

        d_u[d_idx] = d_u0[d_idx] + dtb2 * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2 * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2 * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_w[d_idx]

        # update thermal energy
        d_e[d_idx] = d_e0[d_idx] + dtb2 * d_ae[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z,
               d_u0, d_v0, d_w0, d_u, d_v, d_w, d_e0, d_e, d_au, d_av,
               d_aw, d_ae, dt):

        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_e[d_idx] = d_e0[d_idx] + dt * d_ae[d_idx]



###############################################################################
# `TwoStageRigidBodyStep` class
###############################################################################
class TwoStageRigidBodyStep(IntegratorStep):
    """Simple rigid-body motion

    At each stage of the integrator, the prescribed velocity and
    accelerations are incremented by dt/2.

    Note that the time centered velocity is used for updating the
    particle positions. This ensures exact motion for a constant
    acceleration.

    """
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                   d_u, d_v, d_w, d_u0, d_v0, d_w0):

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

    def stage1(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
               dt):

        dtb2 = 0.5*dt

        d_u[d_idx] = d_u0[d_idx] + dtb2 * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2 * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2 * d_aw[d_idx]

        # positions are updated based on the time centered velocity
        d_x[d_idx] = d_x0[d_idx] + dtb2 * 0.5 * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] = d_y0[d_idx] + dtb2 * 0.5 * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] = d_z0[d_idx] + dtb2 * 0.5 * (d_w[d_idx] + d_w0[d_idx])

    def stage2(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
               dt):

        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        # positions are updated based on the time centered velocity
        d_x[d_idx] = d_x0[d_idx] + dt * 0.5 * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] = d_y0[d_idx] + dt * 0.5 * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] = d_z0[d_idx] + dt * 0.5 * (d_w[d_idx] + d_w0[d_idx])

###############################################################################
# `OneStageRigidBodyStep` class
###############################################################################
class OneStageRigidBodyStep(IntegratorStep):
    """Simple one stage rigid-body motion """
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                   d_u, d_v, d_w, d_u0, d_v0, d_w0):

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

    def stage1(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
                dt):
        pass

    def stage2(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
                dt):

        # update velocities
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_w[d_idx] += dt * d_aw[d_idx]

        # upadte positions using time-centered velocity
        d_x[d_idx] += dt * 0.5 * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] += dt * 0.5 * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] += dt * 0.5 * (d_w[d_idx] + d_w0[d_idx])


###############################################################################
# `VerletSymplecticWCSPHStep` class
###############################################################################
class VerletSymplecticWCSPHStep(IntegratorStep):
    """Symplectic second order integrator described in the review
    paper by Monaghan:

    J. Monaghan, "Smoothed Particle Hydrodynamics", Reports on
    Progress in Physics, 2005, 68, pp 1703--1759 [JM05]

    Notes:

    This integrator should run in PEC mode since in the first stage,
    the positions are updated using the current velocity. The
    accelerations are then computed to advance to the full time step
    values.

    This version of the integrator does not update the density. That
    is, the summation density is used instead of the continuity
    equation.

    """

    def initialize(self):
        pass

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, dt):

        dtb2 = 0.5 * dt

        # Eq. (5.39) in [JM05]
        d_x[d_idx] += dtb2 * d_u[d_idx]
        d_y[d_idx] += dtb2 * d_v[d_idx]
        d_z[d_idx] += dtb2 * d_w[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_ax, d_ay, d_az,
               d_u, d_v, d_w, d_au, d_av, d_aw, dt):

        dtb2 = 0.5 * dt

        # Eq. (5.40) in [JM05]
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_w[d_idx] += dt * d_aw[d_idx]

        # Eq. (5.41) in [JM05] using XSPH velocity correction
        d_x[d_idx] += dtb2 * d_ax[d_idx]
        d_y[d_idx] += dtb2 * d_ay[d_idx]
        d_z[d_idx] += dtb2 * d_az[d_idx]


###############################################################################
# `VelocityVerletSymplecticWCSPHStep` class
###############################################################################
class VelocityVerletSymplecticWCSPHStep(IntegratorStep):
    """Another symplectic second order integrator described in the
    review paper by Monaghan:

    J. Monaghan, "Smoothed Particle Hydrodynamics", Reports on
    Progress in Physics, 2005, 68, pp 1703--1759 [JM05]

    kick--drift--kick form of the verlet integrator

    """

    def initialize(self):
        pass

    def stage1(self, d_idx, d_u, d_v, d_w, d_au, d_av, d_aw, dt):

        dtb2 = 0.5 * dt

        # Eq. (5.51) in [JM05]
        d_u[d_idx] += dtb2 * d_au[d_idx]
        d_v[d_idx] += dtb2 * d_av[d_idx]
        d_w[d_idx] += dtb2 * d_aw[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w,
               d_au, d_av, d_aw, dt):

        dtb2 = 0.5 * dt

        # Eq. (5.52) in [JM05]
        d_x[d_idx] += dt * d_u[d_idx]
        d_y[d_idx] += dt * d_v[d_idx]
        d_z[d_idx] += dt * d_w[d_idx]

        # Eq. (5.53) in [JM05]
        d_u[d_idx] += dtb2 * d_au[d_idx]
        d_v[d_idx] += dtb2 * d_av[d_idx]
        d_w[d_idx] += dtb2 * d_aw[d_idx]

###############################################################################
# `InletOutletStep` class
###############################################################################
class InletOutletStep(IntegratorStep):
    """A trivial integrator for the inlet/outlet particles
    """
    def initialize(self):
        pass

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, dt):
        dtb2 = 0.5*dt
        d_x[d_idx] += dtb2 * d_u[d_idx]
        d_y[d_idx] += dtb2 * d_v[d_idx]
        d_z[d_idx] += dtb2 * d_w[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, dt):
        dtb2 = 0.5*dt
        d_x[d_idx] += dtb2 * d_u[d_idx]
        d_y[d_idx] += dtb2 * d_v[d_idx]
        d_z[d_idx] += dtb2 * d_w[d_idx]



###############################################################################
class LeapFrogStep(IntegratorStep):

    r"""Using this stepper with XSPH as implemented in
    `pysph.base.basic_equations.XSPHCorrection` is not directly possible and
    requires a nicer implementation where the correction alone is added to ``ax,
    ay, az``.
    """

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_ax, d_ay, d_az,
               dt):
        d_x[d_idx] += 0.5 * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += 0.5 * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += 0.5 * dt * (d_w[d_idx] + d_az[d_idx])

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_au, d_v, d_av,
               d_w, d_aw, d_ax, d_ay, d_az,
               d_rho, d_arho, d_e, d_ae, dt):
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_w[d_idx] += dt * d_aw[d_idx]

        d_rho[d_idx] += dt * d_arho[d_idx]
        d_e[d_idx] += dt * d_ae[d_idx]

        d_x[d_idx] += 0.5 * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += 0.5 * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += 0.5 * dt * (d_w[d_idx] + d_az[d_idx])


###############################################################################
class PEFRLStep(IntegratorStep):

    r"""Using this stepper with XSPH as implemented in
    `pysph.base.basic_equations.XSPHCorrection` is not directly possible and
    requires a nicer implementation where the correction alone is added to ``ax,
    ay, az``.
    """

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_ax, d_ay,
               d_az, dt):

        xi = 0.1786178958448091

        d_x[d_idx] += xi * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += xi * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += xi * dt * (d_w[d_idx] + d_az[d_idx])

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_au, d_v, d_av,
               d_w, d_aw, d_ax, d_ay, d_az,
               d_rho, d_arho, d_e, d_ae, dt=0.0):

        lamda = -0.2123418310626054
        fac = (1. - 2.*lamda) / 2.

        d_u[d_idx] += fac * dt * d_au[d_idx]
        d_v[d_idx] += fac * dt * d_av[d_idx]
        d_w[d_idx] += fac * dt * d_aw[d_idx]

        d_rho[d_idx] += fac * dt * d_arho[d_idx]
        d_e[d_idx] += fac * dt * d_ae[d_idx]

        chi = -0.06626458266981849

        d_x[d_idx] += chi * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += chi * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += chi * dt * (d_w[d_idx] + d_az[d_idx])

    def stage3(self, d_idx, d_x, d_y, d_z, d_u, d_au, d_v, d_av,
               d_w, d_aw, d_ax, d_ay, d_az,
               d_rho, d_arho, d_e, d_ae, dt=0.0):

        lamda = -0.2123418310626054

        d_u[d_idx] += lamda * dt * d_au[d_idx]
        d_v[d_idx] += lamda * dt * d_av[d_idx]
        d_w[d_idx] += lamda * dt * d_aw[d_idx]

        d_rho[d_idx] += lamda * dt * d_arho[d_idx]
        d_e[d_idx] += lamda * dt * d_ae[d_idx]

        xi = +0.1786178958448091
        chi = -0.06626458266981849
        fac = 1. - 2.*(xi + chi)

        d_x[d_idx] += fac * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += fac * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += fac * dt * (d_w[d_idx] + d_az[d_idx])

    def stage4(self, d_idx, d_x, d_y, d_z, d_u, d_au, d_v, d_av,
               d_w, d_aw, d_ax, d_ay, d_az,
               d_rho, d_arho, d_e, d_ae, dt=0.0):

        lamda = -0.2123418310626054

        d_u[d_idx] += lamda * dt * d_au[d_idx]
        d_v[d_idx] += lamda * dt * d_av[d_idx]
        d_w[d_idx] += lamda * dt * d_aw[d_idx]

        d_rho[d_idx] += lamda * dt * d_arho[d_idx]
        d_e[d_idx] += lamda * dt * d_ae[d_idx]

        chi = -0.06626458266981849

        d_x[d_idx] += chi * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += chi * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += chi * dt * (d_w[d_idx] + d_az[d_idx])

    def stage5(self, d_idx, d_x, d_y, d_z, d_u, d_au, d_v, d_av,
               d_w, d_aw, d_ax, d_ay, d_az,
               d_rho, d_arho, d_e, d_ae, dt=0.0):

        lamda = -0.2123418310626054
        fac = (1. - 2.*lamda) / 2.

        d_u[d_idx] += fac * dt * d_au[d_idx]
        d_v[d_idx] += fac * dt * d_av[d_idx]
        d_w[d_idx] += fac * dt * d_aw[d_idx]

        d_rho[d_idx] += fac * dt * d_arho[d_idx]
        d_e[d_idx] += fac * dt * d_ae[d_idx]

        xi = +0.1786178958448091

        d_x[d_idx] += xi * dt * (d_u[d_idx] + d_ax[d_idx])
        d_y[d_idx] += xi * dt * (d_v[d_idx] + d_ay[d_idx])
        d_z[d_idx] += xi * dt * (d_w[d_idx] + d_az[d_idx])



import numpy as np
import unittest
import pytest
from pysph.base.particle_array import ParticleArray
import pysph.tools.mesh_tools as G
from pysph.base.utils import get_particle_array


# Data of a unit length cube
def cube_data():
    points = np.array([[0., 0., 0.],
                       [0., 1., 0.],
                       [1., 1., 0.],
                       [1., 0., 0.],
                       [0., 0., 1.],
                       [0., 1., 1.],
                       [1., 0., 1.],
                       [1., 1., 1.]])

    x_cube, y_cube, z_cube = points.T

    cells = np.array([[0, 1, 2],
                      [0, 2, 3],
                      [0, 4, 5],
                      [0, 5, 1],
                      [0, 3, 6],
                      [0, 6, 4],
                      [4, 6, 7],
                      [4, 7, 5],
                      [3, 2, 7],
                      [3, 7, 6],
                      [1, 5, 7],
                      [1, 7, 2]])

    normals = np.array([[0.,  0., -1.],
                        [0.,  0., -1.],
                        [-1.,  0.,  0.],
                        [-1.,  0.,  0.],
                        [0., -1.,  0.],
                        [0., -1.,  0.],
                        [0.,  0.,  1.],
                        [0.,  0.,  1.],
                        [1.,  0.,  0.],
                        [1.,  0.,  0.],
                        [0.,  1.,  0.],
                        [0.,  1.,  0.]])

    vectors = np.zeros((len(cells), 3, 3))
    for i, cell in enumerate(cells):
        idx1, idx2, idx3 = cell

        vector = np.array([[x_cube[idx1], y_cube[idx1], z_cube[idx1]],
                           [x_cube[idx2], y_cube[idx2], z_cube[idx2]],
                           [x_cube[idx3], y_cube[idx3], z_cube[idx3]]])
        vectors[i] = vector

    return x_cube, y_cube, z_cube, cells, normals, vectors


class TestGeometry(unittest.TestCase):
    def test_in_triangle(self):
        assert(G._in_triangle(0.5, 0.5, 0.0, 0.0, 1.5, 0.0, 0.0, 1.5) is True)
        assert(G._in_triangle(1.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0) is False)

    def test_interp_2d(self):
        # Check interpolation between two points on line y=x
        dx = 0.1
        r = G._interp_2d(np.array([0., 0.]), np.array([1., 1.]), dx)
        # Check if all points satisfy y=x
        np.testing.assert_array_almost_equal(
            r[:, 0] - r[:, 1], np.zeros(r.shape[0]))
        # Check if distance between consecutive points is lesser than dx
        np.testing.assert_array_less(np.linalg.norm(r[1:] - r[0:-1], axis=1),
                                     np.ones(r.shape[0] - 1) * dx)

    def test_fill_triangle(self):
        triangle = np.array([[0., 0., 0.],
                             [1., 0., 0.],
                             [0., 1., 0.]])
        dx_triangle = 0.1
        x, y, z = G._fill_triangle(triangle, dx_triangle)
        EPS = np.finfo(float).eps
        np.testing.assert_array_less(-x, np.zeros(x.shape[0]) + EPS)
        np.testing.assert_array_less(-y, np.zeros(x.shape[0]) + EPS)
        np.testing.assert_array_less(-(x + y), np.ones(x.shape[0]) + EPS)
        np.testing.assert_almost_equal(z, np.zeros(x.shape[0]))

    def test_fill_triangle_throws_zero_area_triangle_exception(self):
        self.assertRaises(G.ZeroAreaTriangleException, G._fill_triangle,
                          np.zeros((3, 3)), 0.5)

    def test_fill_triangle_throws_polygon_mesh_error(self):
        self.assertRaises(G.PolygonMeshError, G._fill_triangle,
                          np.zeros((4, 3)), 0.5)

    def test_get_points_from_mgrid(self):
        """Find neighbouring particles around a unit cube"""
        h = 0.1
        x_cube, y_cube, z_cube, cells, normals, vectors = cube_data()

        x, y, z, x_list, y_list, z_list, vectors = \
        G._get_surface_mesh(x_cube, y_cube, z_cube, cells, h, uniform=True)
        pa_mesh = ParticleArray(name='mesh', x=x, y=y, z=z, h=h)
        offset = h
        x_grid, y_grid, z_grid = np.meshgrid(
            np.arange(x.min() - offset, x.max() + offset, h),
            np.arange(y.min() - offset, y.max() + offset, h),
            np.arange(z.min() - offset, z.max() + offset, h))
        pa_grid = ParticleArray(name='grid', x=x_grid, y=y_grid, z=z_grid, h=h)
        x_grid, y_grid, z_grid = G.get_points_from_mgrid(
            pa_grid, pa_mesh, x_list, y_list, z_list, 1, h, vectors, normals
        )

        for i in range(x.shape[0]):
            assert((x[i] ** 2 + y[i] ** 2 + z[i] ** 2) <= 4)

    def _cube_assert(self, x, y, z, h):
        """Check if x,y,z lie within surface of thickness `h` of a unit cube"""
        def surface1(x, y, z): return min(abs(x), abs(1 - x)) < h and \
            y > -h and y < 1 + h and z > -h and z < 1 + h

        def on_surface(x, y, z): return surface1(x, y, z) or \
            surface1(y, x, z) or surface1(z, x, y)

        for i in range(x.shape[0]):
            assert on_surface(x[i], y[i], z[i])

    def test_get_surface_mesh(self):
        """Check if mesh is generated correctly for unit cube"""

        x_cube, y_cube, z_cube, cells, normals, vectors = cube_data()
        x, y, z = G._get_surface_mesh(x_cube, y_cube, z_cube, cells, 0.1)
        h = np.finfo(float).eps
        self._cube_assert(x, y, z, h)

    def test_get_surface_points(self):
        """Check if surface is generated correctly for unit cube"""
        h = 0.1
        x_cube, y_cube, z_cube, cells, normals, vectors = cube_data()
        x, y, z = G.surface_points(x_cube, y_cube, z_cube, cells, h)
        self._cube_assert(x, y, z, h)

    def test_get_surface_points_uniform(self):
        """Check if uniform surface is generated correctly for unit cube"""
        h = 0.1
        x_cube, y_cube, z_cube, cells, normals, vectors = cube_data()
        x, y, z = G.surf_points_uniform(x_cube, y_cube, z_cube,
                                        cells, normals, 1.0, 1.0)
        self._cube_assert(x, y, z, h)

    def test_prism(self):
        tri_normal = np.array([0, -1, 0])
        tri_points = np.array([[0, 0, 0], [1, 0, 0], [0, 0, 1]])
        h = 1/1.5
        prism_normals, prism_points, prism_face_centres = \
            G.prism(tri_normal, tri_points, h)
        assert np.array([-1, 0, 0]) in prism_normals
        assert np.array([0, 1, 0]) in prism_points
        assert np.array([0.5, 0.5, 0]) in prism_face_centres


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from pysph.base.utils import get_particle_array
from pysph.base.nnps import DomainManager
from pysph.sph.basic_equations import SummationDensity
from pysph.tools.sph_evaluator import SPHEvaluator


class TestSPHEvaluator(unittest.TestCase):
    def setUp(self):
        x = np.linspace(0, 1, 10)
        dx = x[1] - x[0]
        self.dx = dx
        m = np.ones_like(x)
        h = np.ones_like(x)*dx
        self.src = get_particle_array(name='src', x=x, m=m, h=h)
        self.equations = [SummationDensity(dest='dest', sources=['src'])]

    def test_evaluation(self):
        # Given
        xd = [0.5]
        hd = self.src.h[:1]
        dest = get_particle_array(name='dest', x=xd, h=hd)
        sph_eval = SPHEvaluator(
            arrays=[dest, self.src], equations=self.equations, dim=1
        )

        # When.
        sph_eval.evaluate()

        # Then.
        self.assertAlmostEqual(dest.rho[0], 9.0, places=2)

    def test_evaluation_with_domain_manager(self):
        # Given
        xd = [0.0]
        hd = self.src.h[:1]
        dest = get_particle_array(name='dest', x=xd, h=hd)
        dx = self.dx
        dm = DomainManager(xmin=-dx/2, xmax=1.0+dx/2, periodic_in_x=True)
        sph_eval = SPHEvaluator(
            arrays=[dest, self.src], equations=self.equations, dim=1,
            domain_manager=dm
        )

        # When.
        sph_eval.evaluate()

        # Then.
        self.assertAlmostEqual(dest.rho[0], 9.0, places=2)

    def test_updating_particle_arrays(self):
        # Given
        xd = [0.5]
        hd = self.src.h[:1]
        dest = get_particle_array(name='dest', x=xd, h=hd)
        sph_eval = SPHEvaluator(
            [dest, self.src], equations=self.equations, dim=1
        )
        sph_eval.evaluate()
        rho0 = dest.rho[0]

        # When.
        dest.x[0] = 0.0
        sph_eval.update_particle_arrays([dest, self.src])
        sph_eval.evaluate()

        # Then.
        self.assertNotEqual(rho0, dest.rho[0])
        self.assertAlmostEqual(dest.rho[0], 7.0, places=1)


if __name__ == '__main__':
    unittest.main()



"""Functions to reduce array data in serial or parallel.
"""

import numpy as np

from cyarray.carray import BaseArray


def _check_operation(op):
    """Raise an exception if the wrong operation is given.
    """
    valid_ops = ('sum', 'max', 'min', 'prod')
    msg = "Unsupported operation %s, must be one of %s."%(op, valid_ops)
    if op not in valid_ops:
        raise RuntimeError(msg)

def _get_npy_array(array_or_carray):
    """Return a numpy array from given carray or numpy array.
    """
    if isinstance(array_or_carray, BaseArray):
        return array_or_carray.get_npy_array()
    else:
        return array_or_carray

def serial_reduce_array(array, op='sum'):
    """Reduce an array given an array and a suitable reduction operation.

    Currently, only 'sum', 'max', 'min' and 'prod' are supported.

    **Parameters**

     - array: numpy.ndarray: Any numpy array (1D).
     - op: str: reduction operation, one of ('sum', 'prod', 'min', 'max')

    """
    _check_operation(op)
    ops = {'sum': np.sum, 'prod': np.prod,
           'max': np.max, 'min': np.min}
    np_array = _get_npy_array(array)
    return ops[op](np_array)


def dummy_reduce_array(array, op='sum'):
    """Simply returns the array for the serial case.
    """
    return _get_npy_array(array)

def mpi_reduce_array(array, op='sum'):
    """Reduce an array given an array and a suitable reduction operation.

    Currently, only 'sum', 'max', 'min' and 'prod' are supported.

    **Parameters**

     - array: numpy.ndarray: Any numpy array (1D).
     - op: str: reduction operation, one of ('sum', 'prod', 'min', 'max')

    """
    np_array = _get_npy_array(array)
    from mpi4py import MPI
    ops = {'sum': MPI.SUM, 'prod': MPI.PROD,
           'max': MPI.MAX, 'min': MPI.MIN}
    return MPI.COMM_WORLD.allreduce(np_array, op=ops[op])

# This is just to keep syntax highlighters happy in editors while writing
# equations.
parallel_reduce_array = mpi_reduce_array



#cython: embedsignature=True

from .nnps_base cimport *

from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector

cimport cython
from cython.operator cimport dereference as deref, preincrement as inc

import numpy as np
cimport numpy as np
from cython.parallel import parallel, prange, threadid

# EPS_MAX is maximum value of eps in tree building
DEF EPS_MAX = 1e-3

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

ctypedef cOctreeNode* node_ptr
ctypedef double* dbl_ptr

cdef extern from *:
    """
    #define START_OMP_PARALLEL_PRAGMA() _Pragma("omp parallel") {
    #define END_OMP_PRAGMA() }
    #define START_OMP_SINGLE_PRAGMA() _Pragma("omp single") {
    #define START_OMP_BARRIER_PRAGMA() _Pragma("omp barrier") {
    """
    void START_OMP_PARALLEL_PRAGMA() nogil
    void END_OMP_PRAGMA() nogil
    void START_OMP_SINGLE_PRAGMA() nogil
    void START_OMP_BARRIER_PRAGMA() nogil

########################################################################

cdef class OctreeNode:
    def __init__(self):
        self.xmin = np.zeros(3, dtype=np.float64)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __richcmp__(self, other, int op):
        # Checks if two nodes are the same.
        # Two nodes are equal if xmin and length are equal.
        # op = 2 corresponds to "=="
        # op = 3 corresponds to "!="
        if type(other) != OctreeNode:
            if op == 2:
                return False
            if op == 3:
                return True
            return NotImplemented

        cdef bint equal_xmin, equal_length
        equal_xmin = True
        equal_length = (self.length == other.length)

        cdef int i
        for i from 0<=i<3:
            if self.xmin[i] != other.xmin[i]:
                equal_xmin = False

        cdef bint equal = equal_xmin and equal_length

        if op == 2:
            return equal
        if op == 3:
            return not equal

        return NotImplemented


    #### Public protocol ################################################

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void wrap_node(self, cOctreeNode* node):
        self._node = node
        self.hmax = node.hmax
        self.length = node.length
        self.is_leaf = node.is_leaf
        self.level = node.level
        self.num_particles = node.num_particles

        self.xmin[0] = self._node.xmin[0]
        self.xmin[1] = self._node.xmin[1]
        self.xmin[2] = self._node.xmin[2]

    cpdef OctreeNode get_parent(self):
        """ Get parent of the node.

        Returns
        -------

        parent : OctreeNode

        """
        if self._node.parent == NULL:
            return None
        cdef OctreeNode parent = OctreeNode()
        parent.wrap_node(self._node.parent)
        return parent

    cpdef UIntArray get_indices(self, Octree tree):
        """ Get indices of a node. Returns empty UIntArray
        if node is not a leaf.

        Returns
        -------

        indices : UIntArray

        """
        if not self._node.is_leaf:
            return UIntArray()
        cdef int idx = self._node.start_index
        cdef UIntArray node_indices = UIntArray()
        cdef u_int* indices = tree.pids
        node_indices.c_set_view(indices + idx,
                self._node.num_particles)
        return node_indices

    cpdef list get_children(self):
        """ Get the children of a node.

        Returns
        -------

        children : list

        """
        if self._node.is_leaf:
            return []
        cdef int i
        cdef list py_children = [None for i in range(8)]
        cdef OctreeNode py_node
        for i from 0<=i<8:
            if self._node.children[i] != NULL:
                py_node = OctreeNode()
                py_node.wrap_node(self._node.children[i])
                py_children[i] = py_node
        return py_children

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef plot(self, ax, color="k"):
        """ Plots a node.

        Parameters
        ----------

        ax : mpl_toolkits.mplot3d.Axes3D instance

        color : color hex string/letter, default ("k")

        """
        cdef int i, j, k
        cdef double x, y, z
        cdef list ax_points = [0,0]

        for i from 0<=i<2:
            for j from 0<=j<2:
                x = self.xmin[0] + i*self.length
                y = self.xmin[1] + j*self.length
                for k from 0<=k<2:
                    ax_points[k] = self.xmin[2] + k*self.length

                ax.plot([x,x], [y,y], zs=ax_points[:], color=color)

        for i from 0<=i<2:
            for k from 0<=k<2:
                x = self.xmin[0] + i*self.length
                z = self.xmin[2] + k*self.length
                for j from 0<=j<2:
                    ax_points[j] = self.xmin[1] + j*self.length

                ax.plot([x,x], ax_points[:], zs=[z,z], color=color)

        for j from 0<=j<2:
            for k from 0<=k<2:
                y = self.xmin[1] + j*self.length
                z = self.xmin[2] + k*self.length
                for i from 0<=i<2:
                    ax_points[i] = self.xmin[0] + i*self.length

                ax.plot(ax_points[:], [y,y], zs=[z,z], color=color)

cdef class Octree:
    def __init__(self, int leaf_max_particles):
        self.leaf_max_particles = leaf_max_particles
        self.depth = 0
        self.root = NULL
        self.leaf_cells = NULL
        self.machine_eps = 16*np.finfo(float).eps
        self.pids = NULL

    def __dealloc__(self):
        if self.root != NULL:
            self._delete_tree(self.root)
        if self.pids != NULL:
            free(self.pids)
        if self.leaf_cells != NULL:
            del self.leaf_cells


    #### Private protocol ################################################

    @cython.cdivision(True)
    cdef inline double _get_eps(self, double length, double* xmin) nogil:
        return (self.machine_eps/length)*fmax(length,
                fmax(fmax(fabs(xmin[0]), fabs(xmin[1])), fabs(xmin[2])))

    @cython.cdivision(True)
    cdef inline void _calculate_domain(self, NNPSParticleArrayWrapper pa_wrapper):
        cdef int num_particles = pa_wrapper.get_number_of_particles()
        pa_wrapper.pa.update_min_max()

        self.xmin[0] = pa_wrapper.x.minimum
        self.xmin[1] = pa_wrapper.y.minimum
        self.xmin[2] = pa_wrapper.z.minimum

        self.xmax[0] = pa_wrapper.x.maximum
        self.xmax[1] = pa_wrapper.y.maximum
        self.xmax[2] = pa_wrapper.z.maximum

        self.hmax = pa_wrapper.h.maximum

        cdef double x_length = self.xmax[0] - self.xmin[0]
        cdef double y_length = self.xmax[1] - self.xmin[1]
        cdef double z_length = self.xmax[2] - self.xmin[2]

        self.length = fmax(x_length, fmax(y_length, z_length))

        cdef double eps = self._get_eps(self.length, self.xmin)

        self.xmin[0] -= self.length*eps
        self.xmin[1] -= self.length*eps
        self.xmin[2] -= self.length*eps

        self.length *= (1 + 2*eps)

    cdef inline cOctreeNode* _new_node(self, double* xmin, double length,
            double hmax = 0, int level = 0, cOctreeNode* parent = NULL,
            int num_particles = 0, bint is_leaf = False) nogil:
        """Create a new cOctreeNode"""
        cdef cOctreeNode* node = <cOctreeNode*> malloc(sizeof(cOctreeNode))

        node.xmin[0] = xmin[0]
        node.xmin[1] = xmin[1]
        node.xmin[2] = xmin[2]

        node.length = length
        node.hmax = hmax
        node.num_particles = num_particles
        node.is_leaf = is_leaf
        node.level = level
        node.start_index = -1

        node.parent = parent

        cdef int i

        for i from 0<=i<8:
            node.children[i] = NULL

        return node

    cdef inline void _delete_tree(self, cOctreeNode* node):
        """Delete octree"""
        cdef int i
        cdef cOctreeNode* temp[8]

        for i from 0<=i<8:
            temp[i] = node.children[i]

        free(node)

        for i from 0<=i<8:
            if temp[i] == NULL:
                continue
            self._delete_tree(temp[i])

    @cython.cdivision(True)
    cdef int _c_build_tree(self, NNPSParticleArrayWrapper pa,
            vector[u_int]* indices, double* xmin, double length,
            cOctreeNode* node, int level) nogil:
        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double xmin_new[3]
        cdef double hmax_children[8]
        cdef int depth_child = 0
        cdef int depth_max = 0

        cdef int i, j, k
        cdef u_int p, q

        for i from 0<=i<8:
            hmax_children[i] = 0

        cdef int oct_id

        # This is required to fix floating point errors. One such case
        # is mentioned in pysph.base.tests.test_octree
        cdef double eps = 2*self._get_eps(length, xmin)

        if (indices.size() < self.leaf_max_particles) or (eps > EPS_MAX):
            copy(indices.begin(), indices.end(), self.pids + self._next_pid)
            node.start_index = self._next_pid
            self._next_pid += indices.size()
            node.num_particles = indices.size()
            del indices
            node.is_leaf = True
            return 1

        cdef vector[u_int]* new_indices[8]
        for i from 0<=i<8:
            new_indices[i] = new vector[u_int]()

        for p from 0<=p<indices.size():
            q = deref(indices)[p]

            find_cell_id_raw(
                    src_x_ptr[q] - xmin[0],
                    src_y_ptr[q] - xmin[1],
                    src_z_ptr[q] - xmin[2],
                    length/2,
                    &i, &j, &k
                    )

            oct_id = k+2*j+4*i

            (new_indices[oct_id]).push_back(q)
            hmax_children[oct_id] = fmax(hmax_children[oct_id], src_h_ptr[q])

        cdef double length_padded = (length/2)*(1 + 2*eps)

        del indices

        for i from 0<=i<2:
            for j from 0<=j<2:
                for k from 0<=k<2:

                    oct_id = k+2*j+4*i

                    if new_indices[oct_id].empty():
                        del new_indices[oct_id]
                        continue

                    xmin_new[0] = xmin[0] + (i - eps)*length/2
                    xmin_new[1] = xmin[1] + (j - eps)*length/2
                    xmin_new[2] = xmin[2] + (k - eps)*length/2

                    eps_new = 2*self._get_eps(length_padded, xmin_new)

                    node.children[oct_id] = self._new_node(xmin_new, length_padded,
                            hmax=hmax_children[oct_id], level=level+1, parent=node)

                    depth_child = self._c_build_tree(pa, new_indices[oct_id],
                            xmin_new, length_padded, node.children[oct_id], level+1)

                    depth_max = <int>fmax(depth_max, depth_child)

        return 1 + depth_max
    
    @cython.cdivision(True)
    @cython.boundscheck(False) 
    cdef int _c_build_tree_level1(self, NNPSParticleArrayWrapper pa, double* xmin, double length,
            cOctreeNode* node, int num_threads) nogil:

        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double xmin_new[3]
        cdef double hmax_children[8]
        cdef cOctreeNode* new_node
        cdef int i, j, k, c, tid, oct_cid, oct_id, count_thread, n, p
        cdef u_int  q
        for i from 0<=i<8:
            hmax_children[i] = 0

        # This is required to fix floating point errors. One such case
        # is mentioned in pysph.base.tests.test_octree
        cdef double eps = 2*self._get_eps(length, xmin)
        n =  node.num_particles

        if (n < self.leaf_max_particles) or (eps > EPS_MAX):
            for i from 0<=i<n:
                self.pids[i] = i
            node.is_leaf = True
            return 1

        cdef u_int* p_indices = self.pids
        cdef vector[u_int]* child_indices = new vector[u_int](n)

        # Number of children of a thread belonging to octant, cumulative_map[thread_id][octant_id]
        cdef vector[vector[int]] cumulative_map = vector[vector[int]](num_threads)
        cdef vector[vector[double]] threads_hmax = vector[vector[double]](num_threads)
        cdef vector[int] count = vector[int](8)

        for i from 0<=i<num_threads:
            cumulative_map[i] = vector[int](8)
            threads_hmax[i] = vector[double](8)

        with nogil, parallel():
            # Required to ensure cython treats i,j,k as private-variables for each thread
            i = j = k = -1
            for p in prange(n,schedule='static'):
                tid = threadid()
                find_cell_id_raw(
                        src_x_ptr[p] - xmin[0],
                        src_y_ptr[p] - xmin[1],
                        src_z_ptr[p] - xmin[2],
                        length/2,
                        &i, &j, &k
                        )

                oct_id = k+2*j+4*i
                deref(child_indices)[p] = oct_id
                cumulative_map[tid][oct_id] += 1
                threads_hmax[tid][oct_id] = fmax(threads_hmax[tid][oct_id], src_h_ptr[p])

        for oct_id from 0<=oct_id<8:
            for tid from 0<=tid<num_threads:
                count_thread = cumulative_map[tid][oct_id]
                cumulative_map[tid][oct_id] = count[oct_id]
                count[oct_id] += count_thread
                hmax_children[oct_id] = fmax(threads_hmax[tid][oct_id], hmax_children[oct_id])

        cdef double length_padded = (length/2)*(1 + 2*eps)
        cdef vector[cOctreeNode *]* next_level_nodes = new vector[node_ptr]()

        c = 0
        for i from 0<=i<2:
            for j from 0<=j<2:
                for k from 0<=k<2:

                    oct_id = k+2*j+4*i

                    if (count[oct_id] == 0):
                        continue

                    xmin_new[0] = xmin[0] + (i - eps)*length/2
                    xmin_new[1] = xmin[1] + (j - eps)*length/2
                    xmin_new[2] = xmin[2] + (k - eps)*length/2

                    eps_new = 2*self._get_eps(length_padded, xmin_new)

                    new_node = self._new_node(xmin_new, length_padded,
                            hmax=hmax_children[oct_id], level=1, parent=node)
                    node.children[oct_id] = new_node
                    new_node.start_index = c
                    new_node.num_particles = count[oct_id]
                    count[oct_id] = c
                    c = c + new_node.num_particles
                    if (new_node.num_particles < self.leaf_max_particles) or (eps_new > EPS_MAX):
                        new_node.is_leaf = True
                        continue
                    next_level_nodes.push_back(new_node)

        # Assign p_indices directly according to which child the index belongs to and
        # where each child of root starts
        with nogil, parallel():
            for p in prange(n, schedule='static'):
                tid = threadid()
                oct_id = deref(child_indices)[p]
                oct_cid = count[oct_id] + cumulative_map[tid][oct_id]
                p_indices[oct_cid] = p
                cumulative_map[tid][oct_id] += 1

        del child_indices

        if (next_level_nodes.empty()):
            del next_level_nodes
            return 2
        else:
            return self._c_build_tree_bfs(pa, p_indices, next_level_nodes, 1, num_threads)

    @cython.cdivision(True)
    @cython.boundscheck(False) 
    cdef int _c_build_tree_bfs(self, NNPSParticleArrayWrapper pa, u_int* p_indices,
             vector[cOctreeNode *]* level_nodes,
            int level, int num_threads) nogil:
        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double eps, eps_new, length_padded, length
        cdef cOctreeNode* node
        cdef int i, j, k, n, num_nodes, l, start, num_p, oct_id, sid, jid, tid, r, d, children
        cdef u_int p, q
        cdef vector[double] hmax_children
        cdef vector[vector[u_int]] new_indices
        cdef double* xmin_new
        cdef double* xmin

        cdef vector[vector[node_ptr]] old_nodes = vector[vector[node_ptr]](num_threads)
        cdef vector[vector[node_ptr]] new_nodes = vector[vector[node_ptr]](num_threads)

        for i from 0<=i<num_threads:
            old_nodes[i] = vector[node_ptr]()
            new_nodes[i] = vector[node_ptr]()

        cdef vector[int] count = vector[int](num_threads+2)
        d = 0
        num_nodes = level_nodes.size()
        for i from 0<=i<num_nodes:
            count[d] = count[d] + 1
            old_nodes[d].push_back(deref(level_nodes)[i])
            d = (d + 1)%num_threads

        # Stores the total nodes that need to be operated on
        count[num_threads] = num_nodes
        # Stores the depth
        count[num_threads + 1] = 2

        del level_nodes

        with nogil, parallel():
            xmin_new = <double *> malloc(3 * sizeof(double))
            xmin = <double *> malloc(3 * sizeof(double))
            hmax_children = vector[double](8)
            new_indices = vector[vector[u_int]](8)

            tid = threadid()

            while(count[num_threads] > 0):
                for i from 0<=i<8:
                    new_indices[i] = vector[u_int]()
                children = 0

                for n from 0<=n<old_nodes[tid].size():
                    for i from 0<=i<8:
                        new_indices[i].clear()
                        hmax_children[i] = 0
                    node = old_nodes[tid][n]
                    length = node.length
                    start = node.start_index
                    num_p = node.num_particles
                    level = node.level
                    xmin[0] = node.xmin[0]
                    xmin[1] = node.xmin[1]
                    xmin[2] = node.xmin[2]

                    for p from start<=p<(start + num_p):
                        q = p_indices[p]
                        find_cell_id_raw(
                            src_x_ptr[q] - xmin[0],
                            src_y_ptr[q] - xmin[1],
                            src_z_ptr[q] - xmin[2],
                            length/2,
                            &i, &j, &k
                            )

                        oct_id = k+2*j+4*i
                        (new_indices[oct_id]).push_back(q)
                        hmax_children[oct_id] = fmax(hmax_children[oct_id], src_h_ptr[q])

                    eps = 2*self._get_eps(length, xmin)
                    length_padded = (length/2)*(1 + 2*eps)


                    for i from 0<=i<2:
                        for j from 0<=j<2:
                            for k from 0<=k<2:
                                oct_id = k+2*j+4*i

                                if new_indices[oct_id].empty():
                                    continue

                                xmin_new[0] = xmin[0] + (i - eps)*length/2
                                xmin_new[1] = xmin[1] + (j - eps)*length/2
                                xmin_new[2] = xmin[2] + (k - eps)*length/2

                                node.children[oct_id] = self._new_node(xmin_new, length_padded,
                                        hmax=hmax_children[oct_id], level=level+1, parent=node)
                                node.children[oct_id].start_index = start
                                num_p = new_indices[oct_id].size()
                                node.children[oct_id].num_particles = num_p

                                # Change the position of the indices in p_indices
                                # according to which child they belong to
                                for l from 0<=l<num_p:
                                    p_indices[start+l] = new_indices[oct_id][l]
                                start = start + new_indices[oct_id].size()

                                eps_new = 2*self._get_eps(length_padded, xmin_new)

                                if (num_p < self.leaf_max_particles) or (eps_new > EPS_MAX):
                                    node.children[oct_id].is_leaf = True
                                    continue
                                new_nodes[tid].push_back(node.children[oct_id])
                                children = children + 1

                count[tid] = children
                old_nodes[tid].clear()

                START_OMP_BARRIER_PRAGMA()
                END_OMP_PRAGMA()

                # Load Balancing is done by single thread

                START_OMP_SINGLE_PRAGMA()
                count[num_threads + 1] = count[num_threads + 1] + 1
                count[num_threads] = 0
                for sid from 0<=sid<num_threads:
                    count[num_threads]  = count[num_threads] + count[sid]
                d =  int(count[num_threads]/num_threads)
                r = count[num_threads]%num_threads
                for sid from 0<=sid<r:
                    count[sid] = d + 1
                for sid from r<=sid<num_threads:
                    count[sid] = d
                END_OMP_PRAGMA()

                old_nodes[tid] = vector[node_ptr](count[tid])

                START_OMP_BARRIER_PRAGMA()
                END_OMP_PRAGMA()

                START_OMP_SINGLE_PRAGMA()
                d = 0
                r = 0
                for sid from 0<=sid<num_threads:
                    for jid from 0<=jid<new_nodes[sid].size():
                        old_nodes[d][r] = new_nodes[sid][jid]
                        r += 1
                        if (r == count[d]):
                            d = d + 1
                            r = 0
                END_OMP_PRAGMA()

                new_nodes[tid].clear()

            free(xmin)
            free(xmin_new)

        return count[num_threads + 1]


    cdef void _plot_tree(self, OctreeNode node, ax):
        node.plot(ax)

        cdef OctreeNode child
        cdef list children = node.get_children()

        for child in children:
            if child == None:
                continue
            self._plot_tree(child, ax)

    cdef void _c_get_leaf_cells(self, cOctreeNode* node):
        if node.is_leaf:
            self.leaf_cells.push_back(node)
            return

        cdef int i
        for i from 0<=i<8:
            if node.children[i] != NULL:
                self._c_get_leaf_cells(node.children[i])


    #### Public protocol ################################################

    cdef int c_build_tree(self, NNPSParticleArrayWrapper pa_wrapper, bint test_parallel = False):

        self._calculate_domain(pa_wrapper)

        cdef int num_particles = pa_wrapper.get_number_of_particles()
        self.num_particles = num_particles

        cdef int i

        cdef vector[u_int]* indices_ptr = new vector[u_int]()

        if self.root != NULL:
            self._delete_tree(self.root)
        if self.pids != NULL:
            free(self.pids)
        if self.leaf_cells != NULL:
            del self.leaf_cells

        self.pids = <u_int*> malloc(num_particles*sizeof(u_int))

        self.root = self._new_node(self.xmin, self.length,
                hmax=self.hmax, level=0)

        cdef int num_threads = get_number_of_threads()

        # Use the parallel method
        if ( (num_threads > 1) or test_parallel ):
            self.root.start_index = 0
            self.root.num_particles = num_particles
            self.depth = self._c_build_tree_level1(pa_wrapper, self.root.xmin,
                    self.root.length, self.root, num_threads)

        # Use the serial method
        else:
            self._next_pid = 0
            for i from 0<=i<num_particles:
                indices_ptr.push_back(i)
            self.depth = self._c_build_tree(pa_wrapper, indices_ptr, self.root.xmin,
                self.root.length, self.root, 0)

        return self.depth

    cdef void c_get_leaf_cells(self):
        if self.leaf_cells != NULL:
            return

        self.leaf_cells = new vector[cOctreeNode*]()
        self._c_get_leaf_cells(self.root)

    @cython.cdivision(True)
    cdef cOctreeNode* c_find_point(self, double x, double y, double z):
        cdef cOctreeNode* node = self.root
        cdef cOctreeNode* prev = self.root

        cdef int i, j, k, oct_id
        while node != NULL:
            find_cell_id_raw(
                    x - node.xmin[0],
                    y - node.xmin[1],
                    z - node.xmin[2],
                    node.length/2,
                    &i, &j, &k
                    )

            oct_id = k+2*j+4*i
            prev = node
            node = node.children[oct_id]

        return prev

    ######################################################################

    cpdef int build_tree(self, ParticleArray pa, bint test_parallel = False):
        """ Build tree.

        Parameters
        ----------

        pa : ParticleArray

        Returns
        -------

        depth : int
        Maximum depth of the tree

        """
        cdef NNPSParticleArrayWrapper pa_wrapper = NNPSParticleArrayWrapper(pa)
        return self.c_build_tree(pa_wrapper, test_parallel)

    cpdef delete_tree(self):
        """ Delete tree"""
        if self.root != NULL:
            self._delete_tree(self.root)
        if self.leaf_cells != NULL:
            del self.leaf_cells
        self.root = NULL
        self.leaf_cells = NULL

    cpdef OctreeNode get_root(self):
        """ Get root of the tree

        Returns
        -------

        root : OctreeNode
        Root of the tree

        """
        cdef OctreeNode py_node = OctreeNode()
        py_node.wrap_node(self.root)
        return py_node

    cpdef list get_leaf_cells(self):
        """ Get all leaf cells in the tree

        Returns
        -------

        leaf_cells : list
        List of leaf cells in the tree

        """
        self.c_get_leaf_cells()
        cdef int i
        cdef list py_leaf_cells = [OctreeNode() for i in range(self.leaf_cells.size())]
        for i from 0<=i<self.leaf_cells.size():
            (<OctreeNode>py_leaf_cells[i]).wrap_node(deref(self.leaf_cells)[i])
        return py_leaf_cells

    cpdef OctreeNode find_point(self, double x, double y, double z):
        """ Get the leaf node to which a point belongs

        Parameters
        ----------

        x, y, z : double
        Co-ordinates of the point

        Returns
        -------

        node : OctreeNode
        Leaf node to which the point belongs

        """
        cdef cOctreeNode* node = self.c_find_point(x, y, z)
        cdef OctreeNode py_node = OctreeNode()
        py_node.wrap_node(node)
        return py_node

    cpdef plot(self, ax):
        """ Plots the tree

        Parameters
        ----------

        ax : mpl_toolkits.mplot3d.Axes3D instance

        """
        cdef OctreeNode root = self.get_root()
        self._plot_tree(root, ax)


cdef class CompressedOctree(Octree):
    def __init__(self, int leaf_max_particles):
        Octree.__init__(self, leaf_max_particles)
        self.dbl_max = np.finfo(float).max

    @cython.cdivision(True)
    cdef int _c_build_tree(self, NNPSParticleArrayWrapper pa,
            vector[u_int]* indices, double* xmin, double length,
            cOctreeNode* node, int level) nogil:

        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double xmin_new[8][3]
        cdef double xmax_new[8][3]
        cdef double length_new = 0
        cdef double hmax_children[8]

        cdef int depth_child = 0
        cdef int depth_max = 0

        cdef int i, j, k
        cdef u_int p, q

        for i from 0<=i<8:
            hmax_children[i] = 0
            for j from 0<=j<3:
                xmin_new[i][j] = self.dbl_max
                xmax_new[i][j] = -self.dbl_max

        cdef int oct_id

        if (indices.size() < self.leaf_max_particles):
            copy(indices.begin(), indices.end(), self.pids + self._next_pid)
            node.start_index = self._next_pid
            self._next_pid += indices.size()
            node.num_particles = indices.size()
            del indices
            node.is_leaf = True
            return 1

        cdef vector[u_int]* new_indices[8]
        for i from 0<=i<8:
            new_indices[i] = new vector[u_int]()

        cdef double* xmin_current
        cdef double* xmax_current

        for p from 0<=p<indices.size():
            q = deref(indices)[p]

            find_cell_id_raw(
                    src_x_ptr[q] - xmin[0],
                    src_y_ptr[q] - xmin[1],
                    src_z_ptr[q] - xmin[2],
                    length/2,
                    &i, &j, &k
                    )

            oct_id = k+2*j+4*i

            (new_indices[oct_id]).push_back(q)

            xmin_current = xmin_new[oct_id]
            xmax_current = xmax_new[oct_id]

            hmax_children[oct_id] = fmax(hmax_children[oct_id], src_h_ptr[q])
            xmin_current[0] = fmin(xmin_current[0], src_x_ptr[q])
            xmin_current[1] = fmin(xmin_current[1], src_y_ptr[q])
            xmin_current[2] = fmin(xmin_current[2], src_z_ptr[q])

            xmax_current[0] = fmax(xmax_current[0], src_x_ptr[q])
            xmax_current[1] = fmax(xmax_current[1], src_y_ptr[q])
            xmax_current[2] = fmax(xmax_current[2], src_z_ptr[q])

        cdef double x_length, y_length, z_length
        cdef double length_padded
        cdef double eps

        del indices

        for i from 0<=i<8:
            if new_indices[i].empty():
                del new_indices[i]
                continue

            xmin_current = xmin_new[i]
            xmax_current = xmax_new[i]

            x_length = xmax_current[0] - xmin_current[0]
            y_length = xmax_current[1] - xmin_current[1]
            z_length = xmax_current[2] - xmin_current[2]

            length_new = fmax(x_length, fmax(y_length, z_length))

            eps = self._get_eps(length_new, xmin_current)

            length_padded = length_new*(1 + 2*eps)

            xmin_current[0] -= length_new*eps
            xmin_current[1] -= length_new*eps
            xmin_current[2] -= length_new*eps

            node.children[i] = self._new_node(xmin_current, length_padded,
                    hmax=hmax_children[i], level=level+1, parent=node)

            depth_child = self._c_build_tree(pa, new_indices[i],
                    xmin_current, length_padded, node.children[i], level+1)

            depth_max = <int>fmax(depth_max, depth_child)

        return 1 + depth_max

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int _c_build_tree_level1(self, NNPSParticleArrayWrapper pa, double* xmin, double length,
            cOctreeNode* node, int num_threads) nogil:

        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double xmin_new[8][3]
        cdef double xmax_new[8][3]
        cdef double hmax_children[8]
        cdef cOctreeNode* new_node
        cdef int i, j, k, c, tid, oct_cid, oct_id, count_thread, n, p
        cdef u_int  q

        for i from 0<=i<8:
            hmax_children[i] = 0
            for j from 0<=j<3:
                xmin_new[i][j] = self.dbl_max
                xmax_new[i][j] = -self.dbl_max


        cdef double eps
        n =  node.num_particles

        if (n < self.leaf_max_particles):
            for i from 0<=i<n:
                self.pids[i] = i
            node.is_leaf = True
            return 1

        cdef u_int* p_indices = self.pids
        cdef vector[u_int]* child_indices = new vector[u_int](n)

        # Number of children of a thread belonging to octant, cumulative_map[thread_id][octant_id]
        cdef vector[vector[int]] cumulative_map = vector[vector[int]](num_threads)
        cdef vector[vector[double]] threads_hmax = vector[vector[double]](num_threads)

        cdef vector[vector[double]] threads_xmin = vector[vector[double]](num_threads)
        cdef vector[vector[double]] threads_ymin = vector[vector[double]](num_threads)
        cdef vector[vector[double]] threads_zmin = vector[vector[double]](num_threads)

        cdef vector[vector[double]] threads_xmax = vector[vector[double]](num_threads)
        cdef vector[vector[double]] threads_ymax = vector[vector[double]](num_threads)
        cdef vector[vector[double]] threads_zmax = vector[vector[double]](num_threads)

        cdef vector[int] count = vector[int](8)

        for i from 0<=i<num_threads:
            cumulative_map[i] = vector[int](8)
            threads_hmax[i] = vector[double](8)

            threads_xmin[i] = vector[double](8)
            threads_ymin[i] = vector[double](8)
            threads_zmin[i] = vector[double](8)

            threads_xmax[i] = vector[double](8)
            threads_ymax[i] = vector[double](8)
            threads_zmax[i] = vector[double](8)

            for j from 0<=j<8:
                threads_xmin[i][j] = self.dbl_max
                threads_ymin[i][j] = self.dbl_max
                threads_zmin[i][j] = self.dbl_max

                threads_xmax[i][j] = -self.dbl_max
                threads_ymax[i][j] = -self.dbl_max
                threads_zmax[i][j] = -self.dbl_max

        with nogil, parallel():
            # Required to ensure cython treats i,j,k as private-variables for each thread
            i = j = k = -1
            for p in prange(n,schedule='static'):
                tid = threadid()
                find_cell_id_raw(
                        src_x_ptr[p] - xmin[0],
                        src_y_ptr[p] - xmin[1],
                        src_z_ptr[p] - xmin[2],
                        length/2,
                        &i, &j, &k
                        )

                oct_id = k+2*j+4*i
                deref(child_indices)[p] = oct_id
                cumulative_map[tid][oct_id] += 1
                threads_hmax[tid][oct_id] = fmax(threads_hmax[tid][oct_id], src_h_ptr[p])

                threads_xmin[tid][oct_id] = fmin(threads_xmin[tid][oct_id], src_x_ptr[p])
                threads_ymin[tid][oct_id] = fmin(threads_ymin[tid][oct_id], src_y_ptr[p])
                threads_zmin[tid][oct_id] = fmin(threads_zmin[tid][oct_id], src_z_ptr[p])

                threads_xmax[tid][oct_id] = fmax(threads_xmax[tid][oct_id], src_x_ptr[p])
                threads_ymax[tid][oct_id] = fmax(threads_ymax[tid][oct_id], src_y_ptr[p])
                threads_zmax[tid][oct_id] = fmax(threads_zmax[tid][oct_id], src_z_ptr[p])

        for oct_id from 0<=oct_id<8:
            for tid from 0<=tid<num_threads:
                count_thread = cumulative_map[tid][oct_id]
                cumulative_map[tid][oct_id] = count[oct_id]
                count[oct_id] += count_thread
                hmax_children[oct_id] = fmax(threads_hmax[tid][oct_id], hmax_children[oct_id])

                xmin_new[oct_id][0] = fmin(threads_xmin[tid][oct_id], xmin_new[oct_id][0])
                xmin_new[oct_id][1] = fmin(threads_ymin[tid][oct_id], xmin_new[oct_id][1])
                xmin_new[oct_id][2] = fmin(threads_zmin[tid][oct_id], xmin_new[oct_id][2])

                xmax_new[oct_id][0] = fmax(threads_xmax[tid][oct_id], xmax_new[oct_id][0])
                xmax_new[oct_id][1] = fmax(threads_ymax[tid][oct_id], xmax_new[oct_id][1])
                xmax_new[oct_id][2] = fmax(threads_zmax[tid][oct_id], xmax_new[oct_id][2])

        cdef double x_length, y_length, z_length, length_new, length_padded
        cdef vector[cOctreeNode *]* next_level_nodes = new vector[node_ptr]()

        cdef double* xmin_current
        cdef double* xmax_current
        c = 0
        for oct_id from 0<=oct_id<8:
            if (count[oct_id] == 0):
                continue

            xmin_current = xmin_new[oct_id]
            xmax_current = xmax_new[oct_id]

            x_length = xmax_current[0] - xmin_current[0]
            y_length = xmax_current[1] - xmin_current[1]
            z_length = xmax_current[2] - xmin_current[2]

            length_new = fmax(x_length, fmax(y_length, z_length))

            eps = self._get_eps(length_new, xmin_current)

            length_padded = length_new*(1 + 2*eps)

            xmin_current[0] -= length_new*eps
            xmin_current[1] -= length_new*eps
            xmin_current[2] -= length_new*eps

            new_node = self._new_node(xmin_current, length_padded,
                    hmax=hmax_children[oct_id], level=1, parent=node)
            node.children[oct_id] = new_node
            new_node.start_index = c
            new_node.num_particles = count[oct_id]
            count[oct_id] = c
            c = c + new_node.num_particles
            if (new_node.num_particles < self.leaf_max_particles):
                new_node.is_leaf = True
                continue
            next_level_nodes.push_back(new_node)

        # Assign p_indices directly according to which child the index belongs to and
        # where each child of root starts
        with nogil, parallel():
            for p in prange(n, schedule='static'):
                tid = threadid()
                oct_id = deref(child_indices)[p]
                oct_cid = count[oct_id] + cumulative_map[tid][oct_id]
                p_indices[oct_cid] = p
                cumulative_map[tid][oct_id] += 1

        del child_indices

        if (next_level_nodes.empty()):
            del next_level_nodes
            return 2
        else:
            return self._c_build_tree_bfs(pa, p_indices, next_level_nodes, 1, num_threads)


    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef int _c_build_tree_bfs(self, NNPSParticleArrayWrapper pa, u_int* p_indices,
             vector[cOctreeNode *]* level_nodes, int level, int num_threads) nogil:

        cdef double* src_x_ptr = pa.x.data
        cdef double* src_y_ptr = pa.y.data
        cdef double* src_z_ptr = pa.z.data
        cdef double* src_h_ptr = pa.h.data

        cdef double eps, length_padded, length, x_length, y_length, z_length, length_new
        cdef cOctreeNode* node
        cdef int i, j, k, n, num_nodes, l, start, num_p, oct_id, sid, jid, tid, r, d, children
        cdef u_int p, q
        cdef vector[double] hmax_children
        cdef vector[vector[u_int]] new_indices
        cdef double *xmin
        cdef vector[dbl_ptr] xmax_new
        cdef vector[dbl_ptr] xmin_new
        cdef double *xmin_current
        cdef double *xmax_current

        cdef vector[vector[node_ptr]] old_nodes = vector[vector[node_ptr]](num_threads)
        cdef vector[vector[node_ptr]] new_nodes = vector[vector[node_ptr]](num_threads)

        for i from 0<=i<num_threads:
            old_nodes[i] = vector[node_ptr]()
            new_nodes[i] = vector[node_ptr]()

        cdef vector[int] count = vector[int](num_threads+2)
        d = 0
        num_nodes = level_nodes.size()
        for i from 0<=i<num_nodes:
            count[d] = count[d] + 1
            old_nodes[d].push_back(deref(level_nodes)[i])
            d = (d + 1)%num_threads
        count[num_threads] = num_nodes
        count[num_threads + 1] = 2

        del level_nodes

        with nogil, parallel():
            xmin = <double *> malloc(3 * sizeof(double))
            hmax_children = vector[double](8)
            new_indices = vector[vector[u_int]](8)
            xmin_new = vector[dbl_ptr](8)
            xmax_new = vector[dbl_ptr](8)

            for i from 0<=i<8:
                xmin_new[i] = <double *> malloc(3 * sizeof(double))
                xmax_new[i] = <double *> malloc(3 * sizeof(double))

            tid = threadid()

            while(count[num_threads]>0):
                for i from 0<=i<8:
                    new_indices[i] = vector[u_int]()
                children = 0

                for n from 0<=n<old_nodes[tid].size():
                    for i from 0<=i<8:
                        new_indices[i].clear()
                        hmax_children[i] = 0
                        for j from 0<=j<3:
                            xmin_new[i][j] = self.dbl_max
                            xmax_new[i][j] = -self.dbl_max

                    node = old_nodes[tid][n]
                    length = node.length
                    start = node.start_index
                    num_p = node.num_particles
                    level = node.level
                    xmin[0] = node.xmin[0]
                    xmin[1] = node.xmin[1]
                    xmin[2] = node.xmin[2]

                    for p from start<=p<(start + num_p):
                        q = p_indices[p]
                        find_cell_id_raw(
                            src_x_ptr[q] - xmin[0],
                            src_y_ptr[q] - xmin[1],
                            src_z_ptr[q] - xmin[2],
                            length/2,
                            &i, &j, &k
                            )

                        oct_id = k+2*j+4*i
                        (new_indices[oct_id]).push_back(q)

                        hmax_children[oct_id] = fmax(hmax_children[oct_id], src_h_ptr[q])

                        xmin_current = xmin_new[oct_id]
                        xmax_current = xmax_new[oct_id]

                        xmin_current[0] = fmin(xmin_current[0], src_x_ptr[q])
                        xmin_current[1] = fmin(xmin_current[1], src_y_ptr[q])
                        xmin_current[2] = fmin(xmin_current[2], src_z_ptr[q])

                        xmax_current[0] = fmax(xmax_current[0], src_x_ptr[q])
                        xmax_current[1] = fmax(xmax_current[1], src_y_ptr[q])
                        xmax_current[2] = fmax(xmax_current[2], src_z_ptr[q])

                    for oct_id from 0<=oct_id<8:
                        if new_indices[oct_id].empty():
                            continue

                        xmin_current = xmin_new[oct_id]
                        xmax_current = xmax_new[oct_id]

                        x_length = xmax_current[0] - xmin_current[0]
                        y_length = xmax_current[1] - xmin_current[1]
                        z_length = xmax_current[2] - xmin_current[2]

                        length_new = fmax(x_length, fmax(y_length, z_length))
                        eps = 2*self._get_eps(length_new, xmin_current)
                        length_padded = length_new*(1 + 2*eps)

                        xmin_current[0] -= length_new*eps
                        xmin_current[1] -= length_new*eps
                        xmin_current[2] -= length_new*eps

                        node.children[oct_id] = self._new_node(xmin_current, length_padded,
                                hmax=hmax_children[oct_id], level=level+1, parent=node)

                        node.children[oct_id].start_index = start
                        num_p = new_indices[oct_id].size()
                        node.children[oct_id].num_particles = num_p

                        # Change the position of the indices in p_indices
                        # according to which child they belong to
                        for l from 0<=l<num_p:
                            p_indices[start+l] = new_indices[oct_id][l]
                        start = start + new_indices[oct_id].size()

                        if (num_p < self.leaf_max_particles):
                            node.children[oct_id].is_leaf = True
                            continue
                        new_nodes[tid].push_back(node.children[oct_id])
                        children = children + 1

                count[tid] = children
                old_nodes[tid].clear()

                START_OMP_BARRIER_PRAGMA()
                END_OMP_PRAGMA()

                # Load Balancing is done by single thread

                START_OMP_SINGLE_PRAGMA()
                count[num_threads + 1] = count[num_threads + 1] + 1
                count[num_threads] = 0
                for sid from 0<=sid<num_threads:
                    count[num_threads]  = count[num_threads] + count[sid]
                d =  int(count[num_threads]/num_threads)
                r = count[num_threads]%num_threads
                for sid from 0<=sid<r:
                    count[sid] = d + 1
                for sid from r<=sid<num_threads:
                    count[sid] = d
                END_OMP_PRAGMA()

                old_nodes[tid] = vector[node_ptr](count[tid])

                START_OMP_BARRIER_PRAGMA()
                END_OMP_PRAGMA()

                START_OMP_SINGLE_PRAGMA()
                d = 0
                r = 0
                for sid from 0<=sid<num_threads:
                    for jid from 0<=jid<new_nodes[sid].size():
                        old_nodes[d][r] = new_nodes[sid][jid]
                        r += 1
                        if (r == count[d]):
                            d = d + 1
                            r = 0
                END_OMP_PRAGMA()

                new_nodes[tid].clear()

            free(xmin)
            for i from 0<=i<8:
                free(xmin_new[i])
                free(xmax_new[i])

        return count[num_threads + 1]

    #### Public protocol ################################################

    cdef int c_build_tree(self, NNPSParticleArrayWrapper pa_wrapper, bint test_parallel = False):

        self._calculate_domain(pa_wrapper)

        cdef int num_particles = pa_wrapper.get_number_of_particles()
        self.num_particles = num_particles

        cdef int i

        cdef vector[u_int]* indices_ptr = new vector[u_int]()

        if self.root != NULL:
            self._delete_tree(self.root)
        if self.pids != NULL:
            free(self.pids)
        if self.leaf_cells != NULL:
            del self.leaf_cells

        self.pids = <u_int*> malloc(num_particles*sizeof(u_int))

        self.root = self._new_node(self.xmin, self.length,
                hmax=self.hmax, level=0)

        cdef int num_threads = get_number_of_threads()

        # Use the parallel method
        if ( (num_threads > 2) or test_parallel ):
            self.root.start_index = 0
            self.root.num_particles = num_particles
            self.depth = self._c_build_tree_level1(pa_wrapper, self.root.xmin,
                    self.root.length, self.root, num_threads)

        # Use the serial method
        else:
            self._next_pid = 0
            for i from 0<=i<num_particles:
                indices_ptr.push_back(i)
            self.depth = self._c_build_tree(pa_wrapper, indices_ptr, self.root.xmin,
                self.root.length, self.root, 0)

        return self.depth



"""Colliding Elastic Balls. (10 minutes)
"""

import numpy

# SPH equations
from pysph.sph.equation import Group
from pysph.sph.solid_mech.basic import (get_particle_array_elastic_dynamics,
                                        ElasticSolidsScheme)

from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.integrator import PECIntegrator
from pysph.sph.integrator_step import SolidMechStep


class Rings(Application):
    def initialize(self):
        # constants
        self.E = 1e7
        self.nu = 0.3975
        self.rho0 = 1.0

        self.dx = 0.0005
        self.hdx = 1.5
        self.h = self.hdx * self.dx

        # geometry
        self.ri = 0.03
        self.ro = 0.04

        self.spacing = 0.041

        self.dt = 1e-8
        self.tf = 5e-5

    def create_particles(self):
        spacing = self.spacing  # spacing = 2*5cm

        x, y = numpy.mgrid[-self.ro:self.ro:self.dx, -self.ro:self.ro:self.dx]
        x = x.ravel()
        y = y.ravel()

        d = (x * x + y * y)
        ro = self.ro
        ri = self.ri
        keep = numpy.flatnonzero((ri * ri <= d) * (d < ro * ro))
        x = x[keep]
        y = y[keep]

        x = numpy.concatenate([x - spacing, x + spacing])
        y = numpy.concatenate([y, y])

        dx = self.dx
        hdx = self.hdx
        m = numpy.ones_like(x) * dx * dx
        h = numpy.ones_like(x) * hdx * dx
        rho = numpy.ones_like(x)

        # create the particle array
        kernel = CubicSpline(dim=2)
        self.wdeltap = kernel.kernel(rij=dx, h=self.h)
        pa = get_particle_array_elastic_dynamics(
            name="solid", x=x + spacing, y=y, m=m,
            rho=rho, h=h, constants=dict(
                wdeltap=self.wdeltap, n=4, rho_ref=self.rho0,
                E=self.E, nu=self.nu))

        print('Ellastic Collision with %d particles' % (x.size))
        print("Shear modulus G = %g, Young's modulus = %g, Poisson's ratio =%g"
              % (pa.G, pa.E, pa.nu))

        u_f = 0.059
        pa.u = pa.cs * u_f * (2 * (x < 0) - 1)

        return [pa]

    def create_scheme(self):
        s = ElasticSolidsScheme(elastic_solids=['solid'], solids=[],
                                dim=2)
        s.configure_solver(dt=self.dt, tf=self.tf, pfreq=500)
        return s


if __name__ == '__main__':
    app = Rings()
    app.run()



"""Evolution of a circular patch of incompressible fluid. (30 seconds)

This is the simplest implementation using existing schemes.

See J. J. Monaghan "Simulating Free Surface Flows with SPH", JCP, 1994, 100, pp
399 - 406
"""
from __future__ import print_function

from numpy import ones_like, mgrid, sqrt

from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme


class EllipticalDrop(Application):
    def initialize(self):
        self.co = 1400.0
        self.ro = 1.0
        self.hdx = 1.3
        self.dx = 0.025
        self.alpha = 0.1

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], [], dim=2, rho0=self.ro, c0=self.co,
            h0=self.dx*self.hdx, hdx=self.hdx, gamma=7.0, alpha=0.1, beta=0.0
        )
        dt = 5e-6
        tf = 0.0076
        s.configure_solver(dt=dt, tf=tf)
        return s

    def create_particles(self):
        """Create the circular patch of fluid."""
        dx = self.dx
        hdx = self.hdx
        ro = self.ro
        name = 'fluid'
        x, y = mgrid[-1.05:1.05+1e-4:dx, -1.05:1.05+1e-4:dx]
        x = x.ravel()
        y = y.ravel()

        m = ones_like(x)*dx*dx*ro
        h = ones_like(x)*hdx*dx
        rho = ones_like(x) * ro
        u = -100*x
        v = 100*y

        # remove particles outside the circle
        indices = []
        for i in range(len(x)):
            if sqrt(x[i]*x[i] + y[i]*y[i]) - 1 > 1e-10:
                indices.append(i)

        pa = get_particle_array(x=x, y=y, m=m, rho=rho, h=h, u=u, v=v,
                                name=name)
        pa.remove_particles(indices)

        print("Elliptical drop :: %d particles"
              % (pa.get_number_of_particles()))

        self.scheme.setup_properties([pa])
        return [pa]


if __name__ == '__main__':
    app = EllipticalDrop()
    app.run()



import numpy as np
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application


class EllipticalDrop(Application):
    def create_particles(self):
        dx = 0.025
        x, y = np.mgrid[-1.05:1.05:dx, -1.05:1.05:dx]
        mask = x*x + y*y < 1
        x = x[mask]
        y = y[mask]
        rho = 1.0
        h = 1.3*dx
        m = rho*dx*dx
        pa = get_particle_array(
            name='fluid', x=x, y=y, u=-100*x, v=100*y, rho=rho,
            m=m, h=h
        )
        return [pa]


if __name__ == '__main__':
    app = EllipticalDrop(fname='ed')
    app.run()



#cython: embedsignature=True

# malloc and friends
from libc.stdlib cimport malloc, free
from libc.math cimport log
from libcpp.vector cimport vector
from libcpp.map cimport map

from cython.operator cimport dereference as deref, preincrement as inc

# Cython for compiler directives
cimport cython

import pyopencl as cl
import pyopencl.array
import pyopencl.algorithm
from pyopencl.scan import GenericScanKernel
from pyopencl.elementwise import ElementwiseKernel

import numpy as np
cimport numpy as np
from mako.template import Template

from pysph.base.gpu_nnps_helper import GPUNNPSHelper
from compyle.array import Array
import compyle.array as array
from compyle.opencl import get_context, get_config, profile_kernel
from compyle.api import Elementwise

from pysph.base.z_order_gpu_nnps_kernels import (ZOrderNbrsKernel,
                                                 ZOrderLengthKernel)

from pysph.base.gpu_helper_kernels import get_elwise, get_scan
from pysph.base.z_order_gpu_nnps_kernels import *

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

    @cython.cdivision(True)
    cdef inline double log2(double n) nogil:
        return log(n)/log(2)


cdef class ZOrderGPUNNPS(GPUNNPS):
    def __init__(self, int dim, list particles, double radius_scale=2.0,
            int ghost_layers=1, domain=None, bint fixed_h=False,
            bint cache=True, bint sort_gids=False, backend=None):
        GPUNNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids, backend
        )

        self.radius_scale2 = radius_scale*radius_scale

        self.src_index = -1
        self.dst_index = -1
        self.sort_gids = sort_gids

        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, num_particles
        self.pids = []
        self.pid_keys = []
        self.cids = []
        self.cid_to_idx = []

        for i from 0<=i<self.narrays:
            pa_wrapper = <NNPSParticleArrayWrapper>self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.pids.append(Array(np.uint32, n=num_particles,
                             backend=self.backend))
            self.pid_keys.append(Array(np.uint64, n=num_particles,
                                 backend=self.backend))
            self.cids.append(Array(np.uint32, n=num_particles,
                             backend=self.backend))
            self.cid_to_idx.append(Array(np.int32,
                                   backend=self.backend))

        self.curr_cid = array.ones(1, dtype=np.uint32,
                                   backend=self.backend)
        self.max_cid_src = array.zeros(1, dtype=np.int32,
                                       backend=self.backend)

        self.dst_to_src = Array(np.uint32, backend=self.backend)
        self.overflow_cid_to_idx = Array(np.int32, backend=self.backend)

        self.cell_lengths = [None] * self.narrays
        self.cell_start_indices = [None] * self.narrays
        self.z_order_nbrs = [None] * 2
        self.z_order_nbr_lengths = [None] * 2

        self.domain.update()
        self.update()

    def get_spatially_ordered_indices(self, int pa_index):
        def update_pids():
            pids_new = array.arange(0, num_particles, 1, dtype=np.uint32,
                                    backend=self.backend)
            self.pids[pa_index].set_data(pids_new)

        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]
        num_particles = pa_wrapper.get_number_of_particles()
        self.sorted = True
        return self.pids[pa_index], update_pids

    cpdef _bin(self, int pa_index):
        self.sorted = False
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]

        fill_pids_knl = get_elwise(fill_pids, self.backend)

        pa_gpu = pa_wrapper.pa.gpu
        fill_pids_knl(pa_gpu.x, pa_gpu.y, pa_gpu.z,
                self.cell_size,
                self.xmin[0], self.xmin[1], self.xmin[2],
                self.pid_keys[pa_index], self.pids[pa_index])

        cdef uint64_t max_key = 1 + array.maximum(self.pid_keys[pa_index])

        self.cell_lengths[pa_index] = array.zeros(max_key, np.int32, 
                                                  backend=self.backend)
        self.cell_start_indices[pa_index] = array.zeros_like(
            self.cell_lengths[pa_index])
        sorted_indices = array.zeros(pa_gpu.get_number_of_particles(), np.int32, 
                                     backend=self.backend)
        sorted_keys = array.zeros(pa_gpu.get_number_of_particles(), np.uint64, 
                                     backend=self.backend)

        count_cells_knl = get_elwise(count_cells, self.backend)
        start_indices_knl = get_scan(input_start_indices, output_start_indices,
                                     np.int32, self.backend)
        sort_knl = get_elwise(sort_indices, self.backend)

        sort_offsets = array.zeros(pa_gpu.get_number_of_particles(), np.int32,
                                   backend=self.backend)

        count_cells_knl(self.pid_keys[pa_index], self.cell_size, 
                        self.cell_lengths[pa_index], sort_offsets)
        start_indices_knl(counts=self.cell_lengths[pa_index], 
                          indices=self.cell_start_indices[pa_index])
        sort_knl(self.pid_keys[pa_index], sort_offsets, 
                 self.cell_start_indices[pa_index],
                 sorted_indices, sorted_keys)

        self.pids[pa_index].set_data(sorted_indices)
        self.pid_keys[pa_index].set_data(sorted_keys)

        self.curr_cid.fill(1)

        fill_unique_cids_knl = get_scan(inp_fill_unique_cids, out_fill_unique_cids,
                                        np.int32, self.backend)

        fill_unique_cids_knl(keys=self.pid_keys[pa_index],
                             cids=self.cids[pa_index])

        curr_cids = self.cids[pa_index]

        cdef unsigned int num_cids = 1 + <unsigned int> curr_cids[-1]
        self.cid_to_idx[pa_index].resize(27 * num_cids)
        self.cid_to_idx[pa_index].fill(-1)

        self.max_cid[pa_index] = num_cids

        map_cid_to_idx_knl= get_elwise(map_cid_to_idx, self.backend)

        map_cid_to_idx_knl(
            pa_gpu.x, pa_gpu.y, pa_gpu.z,
            pa_wrapper.get_number_of_particles(), self.cell_size,
            self.xmin[0], self.xmin[1], self.xmin[2],
            self.pids[pa_index], self.pid_keys[pa_index],
            self.cids[pa_index], self.cid_to_idx[pa_index]
        )

    cpdef _refresh(self):
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, num_particles
        self.max_cid = []
        self.sorted = False

        for i from 0<=i<self.narrays:
            pa_wrapper = <NNPSParticleArrayWrapper>self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.pids[i].resize(num_particles)
            self.pid_keys[i].resize(num_particles)
            self.cids[i].resize(num_particles)
            self.max_cid.append(0)

    cpdef set_context(self, int src_index, int dst_index):
        """Setup the context before asking for neighbors.  The `dst_index`
        represents the particles for whom the neighbors are to be determined
        from the particle array with index `src_index`.

        Parameters
        ----------

         src_index: int: the source index of the particle array.
         dst_index: int: the destination index of the particle array.
        """
        GPUNNPS.set_context(self, src_index, dst_index)

        self.src = self.pa_wrappers[src_index]
        self.dst = self.pa_wrappers[dst_index]

        self.dst_src = src_index != dst_index

        cdef unsigned int overflow_size = 0

        if self.dst_src:
            self.dst_to_src.resize(self.max_cid[dst_index])

            map_dst_to_src_knl = get_elwise(map_dst_to_src, self.backend)

            self.max_cid_src.fill(self.max_cid[src_index])

            map_dst_to_src_knl(self.dst_to_src, self.cids[dst_index],
                    self.cid_to_idx[dst_index],
                    self.pid_keys[dst_index],
                    self.pid_keys[src_index], self.cids[src_index],
                    self.src.get_number_of_particles(), self.max_cid_src)

            overflow_size = <unsigned int>(self.max_cid_src.get()) - \
                    self.max_cid[src_index]

            self.overflow_cid_to_idx.resize(max(1, 27 * overflow_size))
            self.overflow_cid_to_idx.fill(-1)

            fill_overflow_map_knl = get_elwise(fill_overflow_map, self.backend)

            dst_gpu = self.dst.pa.gpu
            fill_overflow_map_knl(self.dst_to_src,
                    self.cid_to_idx[dst_index], dst_gpu.x, dst_gpu.y,
                    dst_gpu.z, self.src.get_number_of_particles(),
                    self.cell_size, self.xmin[0], self.xmin[1], self.xmin[2],
                    self.pid_keys[src_index], self.pids[dst_index],
                    self.overflow_cid_to_idx,
                    np.array(self.max_cid[src_index], dtype=np.uint32))


    cdef void find_neighbor_lengths(self, nbr_lengths):
        if not self.z_order_nbr_lengths[self.dst_src]:
            krnl_source = ZOrderLengthKernel(
                "z_order_nbr_lengths", dst_src=self.dst_src
            )

            self.z_order_nbr_lengths[self.dst_src] = Elementwise(
                krnl_source.function, backend=self.backend
            )

        knl = self.z_order_nbr_lengths[self.dst_src]

        dst_gpu = self.dst.pa.gpu
        src_gpu = self.src.pa.gpu
        knl(dst_gpu.x, dst_gpu.y, dst_gpu.z,
                dst_gpu.h, src_gpu.x, src_gpu.y, src_gpu.z,
                src_gpu.h,
                self.xmin[0], self.xmin[1], self.xmin[2],
                self.src.get_number_of_particles(),
                self.pid_keys[self.src_index],
                self.pids[self.dst_index],
                self.pids[self.src_index],
                self.max_cid[self.src_index], self.cids[self.dst_index],
                self.cid_to_idx[self.src_index],
                self.overflow_cid_to_idx, self.dst_to_src,
                self.radius_scale2, self.cell_size, nbr_lengths)



    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices):
        if not self.z_order_nbrs[self.dst_src]:
            krnl_source = ZOrderNbrsKernel(
                "z_order_nbrs", dst_src=self.dst_src
            )

            self.z_order_nbrs[self.dst_src] = Elementwise(
                krnl_source.function, backend=self.backend
            )

        knl = self.z_order_nbrs[self.dst_src]

        dst_gpu = self.dst.pa.gpu
        src_gpu = self.src.pa.gpu
        knl(dst_gpu.x, dst_gpu.y, dst_gpu.z,
                dst_gpu.h, src_gpu.x, src_gpu.y, src_gpu.z,
                src_gpu.h,
                self.xmin[0], self.xmin[1], self.xmin[2],
                self.src.get_number_of_particles(),
                self.pid_keys[self.src_index],
                self.pids[self.dst_index],
                self.pids[self.src_index],
                self.max_cid[self.src_index], self.cids[self.dst_index],
                self.cid_to_idx[self.src_index],
                self.overflow_cid_to_idx, self.dst_to_src,
                self.radius_scale2, self.cell_size, start_indices, nbrs)



# cython: language_level=3, embedsignature=True
# distutils: language=c++

from libcpp.map cimport map
from libcpp.vector cimport vector

from .nnps_base cimport *

# NNPS using the linked list approach
cdef class LinkedListNNPS(NNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef public IntArray ncells_per_dim  # number of cells in each direction
    cdef public int ncells_tot           # total number of cells
    cdef public bint fixed_h             # Constant cell sizes
    cdef public list heads               # Head arrays for the cells
    cdef public list nexts               # Next arrays for the particles

    cdef NNPSParticleArrayWrapper src, dst # Current source and destination.
    cdef UIntArray next, head              # Current next and head arrays.

    cpdef long _count_occupied_cells(self, long n_cells) except -1
    cpdef long _get_number_of_cells(self) except -1
    cdef long _get_flattened_cell_index(self, cPoint pnt, double cell_size)
    cdef long _get_valid_cell_index(self, int cid_x, int cid_y, int cid_z,
            int* ncells_per_dim, int dim, int n_cells) nogil
    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil
